/* 
 * File:   filter.cpp
 * Author: Aziz Mithani
 *
 * Created on April 7, 2016, 11:39 AM
 */

#include <cstdlib>
#include <string>
#include <fstream>
#include <iostream>
#include "gff.h"
#include "time.h"
#include <time.h>

using namespace std;

static const int FLAG_SEQUENCE_UNMAPPED = 4;
static const int FLAG_SUPPLEMENTARY_ALIGNMENT = 2048;
static const int FLAG_SEQUENCE_ON_REVERSE_STRAND = 16;

void str_split(const string& str, vector<string>& tokens, const string& delimiters) {
    // Skip delimiters at beginning.
    string::size_type last_pos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos = str.find_first_of(delimiters, last_pos);

    while (string::npos != pos || string::npos != last_pos) {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(last_pos, pos - last_pos));
        // Skip delimiters.  Note the "not_of"
        last_pos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, last_pos);
    }
}

bool is_supplementary_alignment(int flag) {
    if (flag & FLAG_SUPPLEMENTARY_ALIGNMENT) {
        return true;
    } else {
        return false;
    }

}

bool is_mapped(int flag) {
    if (flag & FLAG_SEQUENCE_UNMAPPED) {
        return false;
    } else {
        return true;
    }

}

bool is_high_quality(float qual, int mapping_quality_threshold) {
    return (qual < mapping_quality_threshold) ? false : true;
}

bool is_correct_orientation(int flag_1, int flag_2) {

    if ((flag_1 & FLAG_SEQUENCE_ON_REVERSE_STRAND) && (flag_2 & FLAG_SEQUENCE_ON_REVERSE_STRAND)) {//R1R2
        return false;
    } else if (!(flag_1 & FLAG_SEQUENCE_ON_REVERSE_STRAND) && !(flag_2 & FLAG_SEQUENCE_ON_REVERSE_STRAND)) {//F1F2
        return false;
    } else {
        return true;
    }

}

int filter_vcf(int argc, char** argv) {

    string vcf_file = "";
    string out_file = "";
    int snp_quality_threshold = 20;

    // read input arguments
    int i = 2;
    int number_of_arguments = argc - 1;
    while (i < number_of_arguments) {
        string parameter = argv[i++];
        if (parameter.compare("-i") == 0) { // vcf file
            vcf_file = argv[i++];
            continue;
        } else if (parameter.compare("-o") == 0) { // out file
            out_file = argv[i++];
            continue;
        } else if (parameter.compare("-q") == 0) { // snp quality threshold
            snp_quality_threshold = atoi(argv[i++]);
            continue;
        } else {
            cerr << "Invalid option: " << argv[i - 1] << endl;
            return 1;
        }
    }

    // parameter checking
    if (vcf_file.empty()) {
        cerr << "VCF file must be specified." << endl;
        return 1;
    }
    if (out_file.empty()) {
        out_file = vcf_file + ".filt";
    }

    // open the input file stream
    ifstream ifs_vcf(vcf_file.c_str());
    // open the output file 
    ofstream ofs_vcf(out_file.c_str());

    string str_vcf = "";
    while (ifs_vcf.good()) {
        getline(ifs_vcf, str_vcf);

        if (str_vcf.empty()) { // ignore empty lines
            continue;
        } else if (str_vcf[0] == '#') { // write the comments/header
            ofs_vcf << str_vcf << endl;
            continue;
        }

        vector<string> entries;
        str_split(str_vcf, entries, "\t");
        if (entries[4].compare(".") == 0) { // ignore non variant positions
            continue;
        }

        // ignore indels
        if ((entries.size() > 7 && entries[7].find("INDEL") != string::npos) || (entries[3].length() != entries[4].length() && entries[4].find(",") != string::npos)) {
            continue;
        }
        // also ignore rows with more than one base in the reference
        if (entries[3].length() > 1) {
            continue;
        }
        // ignore positions where the reference has an 'N' and the concensus base also has an "N", and does not meets the quality threshold
        if (entries[3].compare("N") == 0) {
            if (entries[4].compare("N") != 0 && entries[4].compare(".") != 0 && atof(entries[5].c_str()) >= snp_quality_threshold && entries[7].find("INDEL") == string::npos) {
                ofs_vcf << str_vcf << endl;
            }
            continue;
        }
        if (atof(entries[5].c_str()) < snp_quality_threshold) {
            continue;
        }

        ofs_vcf << str_vcf << endl;


    } // end while ifsPileup.good()

    ifs_vcf.close();
    ofs_vcf.close();

    return 0;
}

int filter_sam(int argc, char** argv) {

    string sam_file;
    string filtered_sam_file = "";
    string gff_file = "";
    bool paired_data = true;
    int mapping_quality_threshold = 20;
    int max_insert_size = -1; // use -1 for reads to map on same contig; -2 for reads to map on different contigs; +ve value for reads having insert size less than the max_insert_size; 0 to turn off insert size check
    bool both_reads_to_be_mapped = false;

    // read input arguments
    int i = 2;
    int number_of_arguments = argc - 1;
    while (i < number_of_arguments) {
        string parameter = argv[i++];
        if (parameter.compare("-i") == 0) { // SAM file
            sam_file = argv[i++];
            continue;
        } else if (parameter.compare("-o") == 0) { // filtered SAM file
            filtered_sam_file = argv[i++];
            continue;
        } else if (parameter.compare("-g") == 0) { // gff file
            gff_file = argv[i++];
            continue;
        } else if (parameter.compare("-x") == 0) { // max insert size
            max_insert_size = atoi(argv[i++]);
            continue;
        } else if (parameter.compare("-q") == 0) { // mapping quality threshold
            mapping_quality_threshold = atoi(argv[i++]);
            continue;
        } else if (parameter.compare("-p") == 0) { // paired data?
            string str_paired_data = argv[i++];
            paired_data = (str_paired_data[0] == 'T' || str_paired_data[0] == 't' ? true : false);
            continue;
        } else if (parameter.compare("-b") == 0) { // require both reads in a pair to be mapped
            string str_both_reads_to_be_mapped = argv[i++];
            both_reads_to_be_mapped = (str_both_reads_to_be_mapped[0] == 'T' || str_both_reads_to_be_mapped[0] == 't' ? true : false);
            continue;
        } else {
            cerr << "Invalid option: " << argv[i - 1] << endl;
            return 1;
        }

    }

    // parameter checking
    if (sam_file.empty()) {
        cerr << "SAM file must be specified." << endl;
        return 1;
    }

    if (filtered_sam_file.empty()) {
        filtered_sam_file = sam_file + ".filtered";
    }


    // open the input file stream
    ifstream ifs_sam(sam_file.c_str());
    // open the output file stream
    ofstream ofs_filtered(filtered_sam_file.c_str());

    // read gff file, if specified
    gff* gff_ptr;
    map<string, int*> index_array;
    if (!gff_file.empty()) {
        gff_ptr = new gff(gff_file);
        // create index for each gene
        gff_ptr->create_index_array(index_array);
    }


    string read_1 = "";
    string read_2 = "";
    int filtered_reads_count = 0;
    string sequence_name_1 = "";
    string current_sequence_name = ""; // reference (chromosome) being currently read
    //gff::SEQ_GFF_LIST* seq_gff_list_ptr = NULL;
    int* seq_index_array = NULL;
    while (ifs_sam.good()) {
        getline(ifs_sam, read_1);

        if (read_1.empty()) { // ignore empty lines
            continue;
        } else if (read_1[0] == '@') { // header
            // write to the output 
            ofs_filtered << read_1 << endl;
            continue;
        }

        // break read into components
        vector<string> entries_1;
        str_split(read_1, entries_1, "\t");

        int flag_1 = atoi(entries_1[1].c_str());
        float qual_1 = atof(entries_1[4].c_str());

        // ignore supplementary alignment (this will be actually a supplementary alignment of previous read pair - 2nd read)
        if (is_supplementary_alignment(flag_1)) {
            continue;
        }

        // check if this read should be printed
        bool read_1_mapped = is_mapped(flag_1); // read mapped?
        bool is_valid_read = true;
        if (read_1_mapped) { // only check if the read is mapped
            is_valid_read = is_high_quality(qual_1, mapping_quality_threshold); // is mapping quality high?
        }

        bool read_2_mapped = false;
        if (paired_data) {
            // get the 2nd read in the pair
            getline(ifs_sam, read_2);

            // break read into components
            vector<string> entries_2;
            str_split(read_2, entries_2, "\t");
            int flag_2 = atoi(entries_2[1].c_str());
            float qual_2 = atof(entries_2[4].c_str());

            // ignore supplementary alignment (this will be actually a supplementary alignment of the current read pair - 1st read)
            while (is_supplementary_alignment(flag_2)) {
                // get the 2nd read in the pair
                getline(ifs_sam, read_2);

                // break read into components
                entries_2.clear();
                str_split(read_2, entries_2, "\t");
                flag_2 = atoi(entries_2[1].c_str());
                qual_2 = atof(entries_2[4].c_str());
            }

            if (!is_valid_read) { // first read in the pair is not valid, no need to check the second read
                continue;
            }

            // check if this read should be printed
            read_2_mapped = is_mapped(flag_2); // read mapped?

            if (read_2_mapped) {// only check if the read is mapped
                // check if this read should be printed
                is_valid_read = is_high_quality(qual_2, mapping_quality_threshold);
            }

            if (!is_valid_read) { // this read in the pair is not valid, continue
                continue;
            }

            // check for correct orientation. ignore F1F2 and R1R2. Don't check if reads mapped on different contigs/genes/chromosomes is required
            if (max_insert_size == -1) {
                is_valid_read = is_correct_orientation(flag_1, flag_2);
                if (!is_valid_read) { // this read pair is not in the correct orientation
                    continue;
                }
            }

            if (max_insert_size == 0) { // insert size to be ignored when filtering
                is_valid_read = true;
            } else if (max_insert_size > 0 && (abs(atoi(entries_1[8].c_str())) > max_insert_size)) {
                is_valid_read = false;
            } else if (gff_ptr != NULL) {
                // get the sequence name
                sequence_name_1 = entries_1[2];
                string sequence_name_2 = entries_2[2];

                if (sequence_name_1.compare(sequence_name_2) != 0) { // different references, don't print this read pair
                    is_valid_read = (max_insert_size == -1 ? false : true);
                } else {
                    if (sequence_name_1.compare(current_sequence_name) != 0) { // new reference
                        seq_index_array = index_array[sequence_name_1];
                        current_sequence_name = sequence_name_1;
                    }

                    if (seq_index_array == NULL) { // reference not found in the GFF file, ignore 
                        continue;
                    }

                    int read_start_1 = atoi(entries_1[3].c_str());
                    int read_start_2 = atoi(entries_2[3].c_str());


                    if (seq_index_array[read_start_1] == seq_index_array[read_start_2]) {
                        is_valid_read = (max_insert_size == -1 ? true : false);
                    } else {
                        is_valid_read = (max_insert_size == -1 ? false : true);
                    }
                } // if sequence name 1 != sequence name 2
            }

        } else {
            both_reads_to_be_mapped = false;
        }


        if (both_reads_to_be_mapped) {
            if (is_valid_read && read_1_mapped && read_2_mapped) {
                ofs_filtered << read_1 << endl;
                ofs_filtered << read_2 << endl;
                // increment the filtered reads count
                filtered_reads_count += 2;
            }
        } else {
            if (is_valid_read && read_1_mapped) {
                ofs_filtered << read_1 << endl;
                // increment the filtered reads count
                filtered_reads_count++;
            }
            if (is_valid_read && read_2_mapped) {
                ofs_filtered << read_2 << endl;
                // increment the filtered reads count
                filtered_reads_count++;
            }
        }


    } // end while (ifs_sam.good())

    ofs_filtered.close();

    if (!gff_file.empty()) {
        seq_index_array = NULL;
        gff_ptr->delete_index_array(index_array);
        delete gff_ptr;
    }

    return filtered_reads_count;

}

int main(int argc, char** argv) {

    if (argc < 2) {
        cerr << "No option specified!" << endl;
        return -1;
    }

    string command = argv[1];

    if (command.compare("sam") == 0) { // filter sam file
        return filter_sam(argc, argv);
    } else if (command.compare("vcf") == 0) {
        return filter_vcf(argc, argv);
    } else {
        cout << "Invalid option" << endl;
    }


    return 0;
}

