/* 
 * File:   gff.cpp
 * Author: Aiz Mithani
 * 
* Created on April 7, 2016, 11:39 AM
  */

#include <fstream>
#include <iostream>
#include "gff.h"
#include "gff_entry.h"


const string gff::FEATURE_EXON = "exon";
const string gff::FEATURE_GENE = "gene";


gff::gff() {
}

gff::gff(const string& gff_file) {

    // read the gff file
    read_gff_file(gff_file);
}

gff::gff(const gff& old_gff) {
    GFF_LIST::const_iterator the_gff_list = old_gff.gff_list.begin();
    for (; the_gff_list != old_gff.gff_list.end(); the_gff_list++) {
        string sequence_name = the_gff_list->first;

        // allocate the space for GFF entries for this sequence
        SEQ_GFF_LIST* new_seq_gff_list_ptr = new vector<gff_entry*>;

        SEQ_GFF_LIST* seq_gff_list_ptr = the_gff_list->second;
        SEQ_GFF_LIST::iterator the_seq_gff_list = seq_gff_list_ptr->begin();
        for (; the_seq_gff_list != seq_gff_list_ptr->end(); the_seq_gff_list++) {
            // call the copy constructor
            gff_entry* gff_entry_ptr = new gff_entry(**the_seq_gff_list);

            new_seq_gff_list_ptr->push_back(gff_entry_ptr);

        }
        this->gff_list[sequence_name] = new_seq_gff_list_ptr;
    }
}

gff::~gff() {
    GFF_LIST::iterator the_gff_list = this->gff_list.begin();
    for (; the_gff_list != this->gff_list.end(); the_gff_list++) {

        // get the GFF List for this sequence
        SEQ_GFF_LIST* seq_gff_list_ptr = the_gff_list->second;
        // and the iterator
        SEQ_GFF_LIST::iterator the_seq_gff_list = seq_gff_list_ptr->begin();
        // delete GFF entries for all genes
        for (; the_seq_gff_list != seq_gff_list_ptr->end(); the_seq_gff_list++) {
            // delete the GFF entry
            delete *the_seq_gff_list;
        }

        delete seq_gff_list_ptr;
    }

}

void gff::str_split(const string& str, vector<string>& tokens, const string& delimiters) {
    // Skip delimiters at beginning.
    string::size_type last_pos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos = str.find_first_of(delimiters, last_pos);

    while (string::npos != pos || string::npos != last_pos) {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(last_pos, pos - last_pos));
        // Skip delimiters.  Note the "not_of"
        last_pos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, last_pos);
    }
}

gff::SEQ_GFF_LIST* gff::get_gff_list(const string& sequence_name) {
    if (sequence_name.empty()) {
        return NULL;
    } else {
        return this->gff_list[sequence_name];
    }
}

void gff::set_gff_list(const string& sequence_name, SEQ_GFF_LIST* seq_gff_list_ptr) {
    this->gff_list[sequence_name] = seq_gff_list_ptr;
}

void gff::read_gff_file(const string& gff_file) {
    // open the input stream
    ifstream ifs_gff(gff_file.c_str());
    string str_gff = "";
    string previous_sequence_name = "";
    SEQ_GFF_LIST* seq_gff_list_ptr;
    while (ifs_gff.good()) {
        // read the next line from the GFF file
        getline(ifs_gff, str_gff);

        if (str_gff.empty()) { // ignore empty lines
            continue;
        }
        if (str_gff[0] == '#') { // ignore header
            continue;
        }

        vector<string> entries;
        str_split(str_gff, entries, "\t");

        gff_entry* gff_entry_ptr = new gff_entry(entries, str_gff);

        if (gff_entry_ptr->get_sequence_name().compare(previous_sequence_name) != 0) { // new reference is starting
            if (!previous_sequence_name.empty()) { // this is not the first reference
                this->gff_list[previous_sequence_name] = seq_gff_list_ptr;
            }

            // initialise the GFF list for the new sequence
            seq_gff_list_ptr = new vector<gff_entry*>;
            // save the sequence name
            previous_sequence_name = gff_entry_ptr->get_sequence_name();
        }

        seq_gff_list_ptr->push_back(gff_entry_ptr);

    } // end while ifs_gff is good
    this->gff_list[previous_sequence_name] = seq_gff_list_ptr;

}

void gff::create_index_array(map<string, int*>&  index_array) {

    //map<string, int*> index_array;// = new map<string, int*>();
    GFF_LIST::iterator the_gff_list = this->gff_list.begin();
    for (; the_gff_list != this->gff_list.end(); the_gff_list++) {
        string sequence_name = the_gff_list->first;
        SEQ_GFF_LIST* seq_gff_list_ptr = the_gff_list->second;
        
        gff_entry* gff_entry_ptr = *(seq_gff_list_ptr->rbegin());
        int* seq_index_array = new int[gff_entry_ptr->get_end()+1];        
        int gene_idx = 1;
        SEQ_GFF_LIST::iterator the_seq_gff_list = seq_gff_list_ptr->begin();
        for (; the_seq_gff_list != seq_gff_list_ptr->end(); the_seq_gff_list++) {
            gff_entry* gff_entry_ptr = *the_seq_gff_list;
            for (int idx = gff_entry_ptr->get_start() - OFFSET; idx <= gff_entry_ptr->get_end() + OFFSET; idx++) {
                seq_index_array[idx]= gene_idx;
            }
            gene_idx++;
        }
        index_array[sequence_name] = seq_index_array;
    }
    
}

void gff::delete_index_array(map<string, int*>&  index_array) {

    //map<string, int*> index_array;// = new map<string, int*>();
    map<string, int*>::iterator index_array_ptr = index_array.begin();
    for (; index_array_ptr != index_array.end(); index_array_ptr++) {
        delete index_array_ptr->second;
    }

    
}
